package com.sillycycle.bagleyd.ant3d;

/*
 * @(#)Ant3DCanvas.java
 *
 * Copyright 2010 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.Box;
import com.sillycycle.bagleyd.util.Cell3D;
import com.sillycycle.bagleyd.util.CellInfo;
import com.sillycycle.bagleyd.util.Dimension3D;
import com.sillycycle.bagleyd.util.FiniteStateMachine;
import com.sillycycle.bagleyd.util.Graphics3D;
import com.sillycycle.bagleyd.util.Point3D;
import com.sillycycle.bagleyd.util.Turmite3D;

/**
 * The <code>Ant3DCanvas</code> class does all the graphics for
 * <code>Ant3DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant3DCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final byte MAX_STATES = 8;
	static final boolean MONO = false;
	static final String DATAFILE = "data.life";

	Random generator;
	int startSpeed = Ant3DInterface.DEFAULT_SPEED;
	int speed = Ant3DInterface.DEFAULT_SPEED;
	int columns = Ant3DInterface.DEFAULT_COLUMNS;
	int rows = Ant3DInterface.DEFAULT_ROWS;
	int stacks = Ant3DInterface.DEFAULT_STACKS;
	int ncols = columns;
	int nrows = rows;
	int nstacks = stacks;
	int minCol = columns / 2 - 1;
	int maxCol = columns / 2 + 1;
	int minRow = rows / 2 - 1;
	int maxRow = rows / 2 + 1;
	int minStack = stacks / 2 - 1;
	int maxStack = stacks / 2 + 1;
	int area = ncols * nrows;
	int mouseDepth = nstacks / 2;
	boolean randomOrient = false;
	boolean grid = false;
	boolean lissajous = true;
	boolean wrap = true;
	boolean visible = true;
	boolean wireframe = false;
	boolean sphere = false;
	boolean useTable = false;
	boolean mouth = true;
	//boolean focus = true;
	boolean mouseDown = false;
	boolean edge = false;
	int neighbors =
		Ant3DRules.indexToNeighbor[Ant3DInterface.DEFAULT_NEIGHBOR_INDEX];
	Color initialCellColor = Color.red;
	Color gridColor = Color.gray;
	Color background = Color.black;
	Cell3D cellPl;
	transient Graphics3D g3d;
	transient Capturer capturer;
	JFrame frame = null;
	boolean debug = false;
	String name = "newbie";
	Ant3DMatrix field = null;
	int gen = 0;
	int nCells = 0, cCells = 0;
	String stringSave = null;
	transient Ant3DRules param = null;
	transient Turmite3D turmite = null;
	int rule = Ant3DRules.MIN_RULE;

	private synchronized void pulse(boolean draw) {
		int tapePos, newPosition, statePos, col, row, stack;
		byte color, newColor;
		short oldDir, oldGrav, arrow, newDir, newGrav;
		FiniteStateMachine status;
		Turmite3D newTurmite;

		if (lissajous)
			g3d.lissajous();
		cCells = 0;
		if (draw)
			eraseFrame(); // need for 3D as erase cell is hard
		col = turmite.getColumn();
		row = turmite.getRow();
		stack = turmite.getStack();
		tapePos = getPosition(col, row, stack, ncols, area);
		color = field.getCell(tapePos); /* read tape */
		statePos = color + turmite.getState() * param.getNColors();
		status = param.machine[statePos];
		newColor = status.getColor();
		if (debug)
			System.out.println("statePos " + statePos +
				", newColor " + newColor);
		// Drawn later so it will look 3D
		//CellInfo info = new CellInfo(tapePos,
		//	newColor, false);
		//drawCell(info);
		field.setCell(tapePos, newColor); /* write on tape, ncells updated */
		checkAliveBounds(col, row, stack);
		/* Find direction of Bees or Ants. */
		/* Translate relative direction to actual direction */
		oldDir = turmite.getDirection();
		oldGrav = turmite.getGravity();
		arrow = status.getDirection();
		newTurmite = heading(oldDir, oldGrav, arrow);
		newDir = newTurmite.getDirection();
		newGrav = newTurmite.getGravity();
		turmite.setDirection(newDir);
		turmite.setGravity(newGrav);
		//truchet stuff goes here
		turmite.setState(status.getNext());

		/* Allow step first then turn */
		//System.out.println("before:" + oldDir + " " + arrow);
		oldDir = ((arrow < neighbors) ? newDir : oldDir);
		//System.out.println("after:" + oldDir + " " + arrow);
		if (debug)
			System.out.println("newDir " + newDir +
				", newGrav " + newGrav +
				", oldDir " + oldDir +
				", col " + col +
				", row " + row +
				", stack " + stack);
		newPosition = positionOfNeighbor(oldDir, col, row, stack);
		int newCol = getColumn(newPosition);
		int newRow = getRow(newPosition);
		int newStack = getStack(newPosition);
		turmite.setPosition(newCol, newRow, newStack);
		field.setListCell(newPosition, MAX_STATES);
		if (debug)
			System.out.println("neighbors " +
				neighbors + ", newDir " + newDir +
				", newCol " + newCol +
				", newRow " + newRow +
				", newStack " + newStack);
		if (draw)
			redrawField();
		// Do later (cannot write this to tape)
		//drawTurmite(newDir, newGrav, newCol, newRow, newStack);
		gen++;
	}

	void turmitePulse(boolean draw, boolean fast) {
		if (lissajous)
			g3d.lissajous();
		pulse(draw);
		if (draw && (!fast || (gen & 63) == 0)) {
			((Ant3DFrame) frame).updateGen(gen);
			((Ant3DFrame) frame).updateCells(nCells);
		}
	}

	public Ant3DCanvas() {
		generator = new Random(System.nanoTime());
		capturer = new Capturer();
		inits();
	}

	final void inits() {
		initTable();
		resizeField();
		clearField(true);
		prepopulate();
	}

	void initTable() {
		param = new Ant3DRules(neighbors/*,
			Ant3DRules.getRuleName(getNeighborIndex())*/);
		param.initMachine();
		param.parseRule(neighbors, rule, useTable);
	}

	void ruleChange() {
		if ((Ant3DFrame) frame != null) {
			param = new Ant3DRules(neighbors/*,
				Ant3DRules.getRuleName(getNeighborIndex())*/);
		}
	}

	public int getNeighborIndex() {
		return Ant3DRules.neighborToIndex(neighbors);
	}

	static boolean polyhedronChange(int oldNeighborIndex, int newNeighborIndex) {
		if (oldNeighborIndex != newNeighborIndex) {
			int oldPolyhedron = Ant3DRules.neighborIndexToPolyhedron[oldNeighborIndex];
			int newPolyhedron = Ant3DRules.neighborIndexToPolyhedron[newNeighborIndex];

			return (oldPolyhedron != newPolyhedron);
		}
		return false;
	}

	int getPolyhedron() {
		return Ant3DRules.neighborToPolyhedron(neighbors);
	}

	void setNeighbors(int neighborIndex) {
		neighbors = Ant3DRules.indexToNeighbor[neighborIndex];
	}

	int getRealColumns(int cols) {
		int polyhedron = getPolyhedron();
		int realColumns = cols;

		if (polyhedron == 14 && (realColumns & 1) == 1) {
			realColumns--;
		}
		return realColumns;
	}

	int getStacksFromColumns(int cols) {
		int polyhedron = getPolyhedron();
		int calculatedStacks = cols / 2;
		//int calculatedStacks = columns;

		if (!boundsCheck12(polyhedron, cols, cols, calculatedStacks)) {
			calculatedStacks++;
		} else if (!boundsCheck14(polyhedron, calculatedStacks)) {
			calculatedStacks = cols;
		}
		return calculatedStacks;
	}

	public int getColumns() {
		return this.ncols;
	}

	public int getRows() {
		return this.nrows;
	}

	public int getStacks() {
		return this.nstacks;
	}

	/* Checks to make sure there is a cell there */
	static boolean cellCheck(int polyhedron, int x, int y, int z) {
		return (polyhedron != 12 || ((x + y + z) & 1) == 0);
	}

	/* Checks to make sure it wraps well */
	static boolean boundsCheck12(int polyhedron, int x, int y, int z) {
		return (polyhedron != 12 || ((x + y + z) & 1) == 0);
	}

	static boolean boundsCheck14(int polyhedron, int z) {
		return (polyhedron != 14 || (z & 1) == 0);
	}

	public void speedUp() {
		speed += 10;
		if (speed > Ant3DInterface.MAX_SPEED)
			speed = Ant3DInterface.MAX_SPEED;
		((Ant3DFrame) frame).setDelay(Ant3DInterface.MAX_SPEED +
			Ant3DInterface.MIN_SPEED - 1 - speed);
	}

	public void slowDown() {
		speed -= 10;
		if (speed < Ant3DInterface.MIN_SPEED)
			speed = Ant3DInterface.MIN_SPEED;
		((Ant3DFrame) frame).setDelay(Ant3DInterface.MAX_SPEED +
			Ant3DInterface.MIN_SPEED - 1 - speed);
	}

	void resizeField() {
		int newWidth = getSize().width;
		int newHeight = getSize().height;
		int tempCols = ncols;
		int oldArea = area;
		int polyhedron = getPolyhedron();
		int newI = 0, newJ = 0, newK = 0;

		capturer.clearCapture();
		if ((Ant3DFrame) frame != null) {
			tempCols = Ant3DFrame.getColumns();
		}
		ncols = getRealColumns(tempCols);
		nrows = ncols;
		area = ncols * nrows;
		nstacks = getStacksFromColumns(tempCols);
		mouseDepth = nstacks / 2;
		if (newWidth == 0)
			newWidth = Ant3DInterface.DEFAULT_CELLSIZE *
				ncols + 4;
		if (newHeight == 0)
			newHeight = Ant3DInterface.DEFAULT_CELLSIZE *
				nrows + 4;
		cellPl = new Cell3D(polyhedron, ncols, nrows/*, nstacks*/,
			newWidth, newHeight);
//System.out.println("Resize " + neighbors);
//System.out.println("ncols " + ncols + ", " + nrows);
//System.out.println("cols " + columns + ", " + rows);
//System.out.println(param.getCommonName());
		if (field == null) {
			columns = ncols;
			rows = nrows;
			stacks = nstacks;
			field = new Ant3DMatrix(ncols, nrows, nstacks);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			minStack = nstacks / 2 - 1;
			maxStack = nstacks / 2 + 1;
			nCells = 0;
			if ((Ant3DFrame) frame != null) {
				((Ant3DFrame) frame).initPopup();
			}
		} else if (columns != ncols || rows != nrows ||
				stacks != nstacks) {
			if (!boundsCheck12(polyhedron, minCol, minRow, minStack)) {
				if (minStack > 0)
					minStack--;
				else if (minRow > 0)
					minRow--;
				else if (minCol > 0)
					minCol--;
			} else if (!boundsCheck14(polyhedron, minStack)) {
				minStack--;
			}

			int fieldX = maxCol - minCol + 1;
			int fieldY = maxRow - minRow + 1;
			int fieldZ = maxStack - minStack + 1;
			int fieldArea = fieldX * fieldY;
			Ant3DMatrix cage = new Ant3DMatrix(fieldX, fieldY, fieldZ);
			for (CellInfo cellInfo : field.list) {
				int oldP = cellInfo.getPosition();
				int i = getColumn(oldP, columns);
				int j = getRow(oldP, columns, oldArea);
				int k = getStack(oldP, oldArea);
				byte value = field.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i - minCol, j - minRow,
						k - minStack, fieldX, fieldArea);
					cage.setArrayCell(p, value);
				}
			}
			if (turmite != null) {
				newI = turmite.getColumn() - minCol;
				newJ = turmite.getRow() - minRow;
				newK = turmite.getStack() - minStack;
			}
			field.clearLiveCells();
			columns = ncols;
			rows = nrows;
			stacks = nstacks;
			field = new Ant3DMatrix(ncols, nrows, nstacks);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			minStack = nstacks / 2 - 1;
			maxStack = nstacks / 2 + 1;
			nCells = 0;

			int offsetX = minCol + 1 - fieldX / 2;
			int offsetY = minRow + 1 - fieldY / 2;
			int offsetZ = minStack + 1 - fieldZ / 2;

			if (!boundsCheck12(polyhedron, offsetX, offsetY, offsetZ)
					|| !boundsCheck14(polyhedron, offsetZ)) {
				offsetZ--;
			}
			for (int k = 0; k < fieldZ; k++)
			if ((k + offsetZ >= 0) && (k + offsetZ <= nstacks - 1))
			for (int j = 0; j < fieldY; j++)
			if ((j + offsetY >= 0) && (j + offsetY <= nrows - 1))
			for (int i = 0; i < fieldX; i++)
			if ((i + offsetX >= 0) && (i + offsetX <= ncols - 1)) {
				int oldP = getPosition(i, j, k,
					fieldX, fieldArea);
				byte value = cage.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i + offsetX, j + offsetY, k + offsetZ,
						ncols, area);
					field.setCell(p, value);
					nCells++;
				}
			}
			if (turmite != null) {
				turmite.setPosition(newI + offsetX, newJ + offsetY, newK + offsetZ);
			}
			if ((Ant3DFrame) frame != null) {
				((Ant3DFrame) frame).updateGen(gen);
				((Ant3DFrame) frame).updateCells(nCells);
				((Ant3DFrame) frame).initPopup();

			}
		}
		g3d = new Graphics3D(this.getGraphics(),
			ncols, nrows, nstacks,
			newWidth, newHeight, mouseDepth);
	}

	void xorCell(CellInfo info, int k) {
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);
		byte state = info.getCell();
		Color color;

		if (state == 0) {
			color = background;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorWheel(state);
			}
		}
		if ((k & 1) == 0) // Simulate Xor
			color = new Color(255 - color.getRed(),
				255 - color.getGreen(),
				255 - color.getBlue());
		cellPl.drawCell(g3d, wireframe, sphere,
			color, gridColor, col, row, stack);
	}

	void showLiveCells() {
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			cellInfo.setDistance(g3d.getDistance(getColumn(position), getRow(position), getStack(position)));
		}
		/* Sort live cells and display. */
		java.util.Collections.sort(field.list, new Ant3DMatrix());
		int col = turmite.getColumn();
		int row = turmite.getRow();
		int stack = turmite.getStack();
		short direction = turmite.getDirection();
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int i = getColumn(position);
			int j = getRow(position);
			int k = getStack(position);
			/*System.out.println("distance: " +
			i + " " + j + " " + k + " : " +
			cellInfo.getDistance());*/
			int state = field.getCell(position);
			if (state != 0) {
				drawCell(cellInfo);
				nCells++;
				checkAliveBounds(i, j, k);
			}
			if (i == col && j == row && k == stack) {
				CellInfo info = new CellInfo(
					position, MAX_STATES, (byte) 0);
				if (neighbors == 8)
					direction = -1;
				drawCellMouth(info, direction);
			}
		}
	}

	// Should make some attempt at minimizing size if on a border...
	void printField(StringBuilder sb) {
		String space = " ";
		String newline = NL;
		String pound = "#";
		boolean firstPoint = true;
		int i, j, k, p;
		int startX = 0, startY = 0, startZ = 0;
		int endX = 0, endY = 0, endZ = 0;
		int polyhedron = getPolyhedron();

		sb.append(pound).append("Dim").append(space)
			.append(3).append(newline);
		sb.append(pound).append("Neighbors").append(space)
			.append(neighbors).append(newline);
		sb.append(pound).append("Rule").append(space)
			.append(rule).append(newline);
		sb.append(pound).append("Cells").append(space)
			.append(nCells).append(newline);
		sb.append(pound).append("Name").append(space)
			.append(name).append(newline);
		/*sb.append(pound).append("Sig").append(space)
			.append("An A0 <> D1 Dn").append(newline);*/
		for (k = 0; k < nstacks; k++)
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++) {
				p = getPosition(i, j, k);
				if (field.getCell(p) != 0) {
					if (firstPoint) {
						startX = i;
						startY = j;
						startZ = k;
						endX = i;
						endY = j;
						endZ = k;
						firstPoint = false;
					} else {
						if (i < startX)
							startX = i;
						if (j < startY)
							startY = j;
						if (k < startZ)
							startZ = k;
						if (i > endX)
							endX = i;
						if (j > endY)
							endY = j;
						if (k > endZ)
							endZ = k;
					}
				}
		}
		if (polyhedron == 12 && ((startX + startY + startZ) & 1) == 1) {
			if (startZ >= startX && startZ >= startY)
				startZ--;
			else if (startY >= startZ && startY >= startX)
				startY--;
			else
				startX--;
		} else if (polyhedron == 14 && ((startZ) & 1) == 1) {
			startZ--;
		}
		sb.append(pound).append("Width").append(space)
			.append(endX - startX).append(1).append(newline);
		sb.append(pound).append("Height").append(space)
			.append(endY - startY).append(1).append(newline);
		sb.append(pound).append("Depth").append(space)
			.append(endZ - startZ).append(1).append(newline);
		sb.append(pound).append("P").append(space)
			.append(startX - ncols / 2).append(space).append(startY - nrows / 2).append(space)
			.append(startZ - nstacks / 2).append(newline);
		/*sb.append(pound).append("P").append(space)
			.append(0).append(space).append(0).append(newline);*/
		for (k = startZ; k <= endZ; k++) {
		for (j = startY; j <= endY; j++) {
			if (polyhedron == 14 && (k & 1) == 1)
				sb.append(space);
		for (i = startX; i <= endX; i++) {
				p = getPosition(i, j, k);
				if (polyhedron == 12 && ((i + j + k) & 1) == 0) {
					sb.append(0);
				} else {
					sb.append(field.getCell(p));
				}
				if (polyhedron == 14)
					sb.append(space);
		}
		sb.append(newline);
		}
		sb.append(newline);
		}
	}

	void clearField(boolean draw) {
		field.clearLiveCells();
		if (draw)
			eraseFrame();
		gen = 0;
		edge = false;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		minStack = nstacks / 2 - 1;
		maxStack = nstacks / 2 + 1;
		nCells = 0;
		if (draw && (Ant3DFrame) frame != null) {
			//Graphics g = this.getGraphics();

			((Ant3DFrame) frame).updateGen(gen);
			((Ant3DFrame) frame).updateCells(nCells);
			//drawFrame(g, true);
			if (grid)
				drawGrid();
		}
	}

	void prepopulate() {
		int col = ncols / 2;
		int row = nrows / 2;
		int stack = nstacks / 2;
		int polyhedron = getPolyhedron();

		if (!cellCheck(polyhedron, col, row, stack)) {
			col--;
		}
		turmite = new Turmite3D(col, row, stack, (byte) 0, (byte) 5,
			(byte) 0);
		field.setListCell(getPosition(col, row, stack), MAX_STATES);
		//drawTurmite((short) 0, (short) 5, col, row, stack);
		//redrawField();
	}

	void showMessage(String msg) {
		((Ant3DFrame) frame).showMessage(msg);
	}

	void setCell(int col, int row, int stack) {
		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows &&
				stack >= 0 && stack < nstacks) {
			setCell(col, row, stack, (byte) 1);
		}
	}

	private void setOrientCell(int x, int y, int z,
			int startX, int startY, int startZ,
			byte state, int orient) {
		int i = x, j = y, k = z, temp;
		int polyhedron = getPolyhedron();

		if (orient >= 16) {
			temp = i;
			i = j;
			if (polyhedron == 14) {
				j = (k >= 0) ? (k / 2) : (k - 1) / 2;
				k = (k & 1) + 2 * temp;
			} else {
				j = k;
				k = temp;
			}
		} else if (orient >= 8) {
			temp = i;
			if (polyhedron == 14) {
				i = (k >= 0) ? (k / 2) : (k - 1) / 2;
				k = (k & 1) + 2 * j;
			} else {
				i = k;
				k = j;
			}
			j = temp;
		}
		if (orient % 8 >= 4) {
			if (polyhedron == 14 && (k & 1) == 1)
				i = -i - 1;
			else
				i = -i;
		}
		if (orient % 4 >= 2) {
			if (polyhedron == 14 && (k & 1) == 1)
				j = -j - 1;
			else
				j = -j;
		}
		if ((orient & 1) == 1) {
			k = -k;
		}
		i += startX;
		j += startY;
		k += startZ;
		// do not draw cells here as they may be in wrong depth order
		if (i >= 0 && i < ncols && j >= 0 && j < nrows &&
				k >= 0 && k < nstacks)
			setCell(i, j, k, state);
	}

	public void displayForm(int[] data, boolean draw) {
		int n = 0, i = 0, j = 0, k;
		int startX, startY, startZ;
		int orient = 0;
		int polyhedron = getPolyhedron();

		startX = ncols / 2;
		startY = nrows / 2;
		startZ = nstacks / 2;
		if (!boundsCheck12(polyhedron, startX, startY, startZ)
				|| !boundsCheck14(polyhedron, startZ))
			startZ++;
		if (randomOrient)
			orient = generator.nextInt(24);
		while (n < data.length) {
			switch (n % 3) {
			case 2:
			    k = data[n];
			    if (cellCheck(polyhedron, i, j, k)) {
				setOrientCell(i, j, k,
					startX, startY, startZ,
					(byte) 1, orient);
			    }
			    break;
			case 1:
			    j = data[n];
			    break;
			default:
			    i = data[n];
			    break;
			}
			n++;
		}
		if (draw)
			redrawField();
	}

	public void getAnt() {
		this.stringSave = ((Ant3DFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getAntString();
	}

	public void getAntFile() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder fileData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				fileData.append(lineData).append(NL);
			this.stringSave = fileData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file to read");
				return;
			}
			getAntString();
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getAntString() {
		try {
			String space = " ";
			String newline = NL;
			String pound = "#";
			int dim = 0, neighborhood, cells;
			int width = 5, height = 5, depth = 5;
			int offsetX = 0, offsetY = 0, offsetZ = 0;
			int i = 0;
			String ruleString;
			String[] lineToken = this.stringSave.split(newline);
			if (lineToken.length == 0)
				return;
			String lineData = lineToken[i];
			String[] wordToken = null;

			try {
				while (i < lineToken.length) {
					wordToken = lineData.split(space);
					if (wordToken.length > 1 &&
							(pound + "Dim").equals(wordToken[0]))
						break;
					if (++i == lineToken.length)
						return;
					lineData = lineToken[i];
				}
				if (wordToken != null && wordToken.length > 1)
					dim = Integer.parseInt(wordToken[1]);
				if (dim != 3) {
					showMessage(DATAFILE +
						" corrupted: dim " + dim);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: dim not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Neighbors
				neighborhood = Integer.parseInt(wordToken[1]);
				if (Math.abs(neighborhood) <= 2) {
					showMessage(DATAFILE +
						" corrupted: neighbors " + neighborhood);
					neighborhood = this.neighbors;
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: neighbors not an integer");
				neighborhood = this.neighbors;
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #Rule
			ruleString = Ant3DRules.parse(/*neighbors, */wordToken[1]);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Cells
				cells = Integer.parseInt(wordToken[1]);
				if (cells < 0) {
					showMessage(DATAFILE +
						" corrupted: cells " + cells);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: cells not an integer");
			}
			lineData = lineToken[++i];
			// #Name
			name = lineData.substring(lineData.indexOf(space) + 1);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Width
				width = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: width not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Height
				height = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: height not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Depth
				depth = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: depth not an integer");
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #P
			if (wordToken.length > 1)
				offsetX = Integer.parseInt(wordToken[1]);
			if (wordToken.length > 2)
				offsetY = Integer.parseInt(wordToken[2]);
			if (wordToken.length > 3)
				offsetZ = Integer.parseInt(wordToken[3]);
			//((Ant3DFrame) frame).callback(
			//	Ant3DInterface.ACTION_RESTORE);
			cells = scanField(lineToken, i, offsetX, offsetY, offsetZ);
			drawField(true);
			System.out.println(DATAFILE +
				": neighbors " + neighborhood +
				", rule " + ruleString +
				", cells " + cells +
				", name " + name +
				", width " + width +
				", height " + height +
				", depth " + depth +
				", offsetX " + offsetX +
				", offsetY " + offsetY +
				", offsetZ " + offsetZ);
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading");
		}
	}

	public void writeAnt() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		((Ant3DFrame) frame).writeTextArea(sb.toString());
	}

	public void writeAntFile() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.println(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public void printAnt() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		System.out.println(sb);
	}

	void eraseFrame(Graphics g) {
		g.setColor(background);
		g.fillRect(0, 0, getSize().width, getSize().height);
	}

	void eraseFrame() {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			eraseFrame(g);
		} finally {
			g.dispose();
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	/*void drawFrame(Graphics g, boolean show) {
		Box b = cellPl.boundBox(ncols, nrows, nstacks);

		if (g == null || b == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, b.y);
		g.fillRect(0, b.y, b.x, b.height - b.y);
		g.fillRect(0, b.height, getSize().width, b.y);
		g.fillRect(b.width, b.y,
			getSize().width - b.width, b.height - b.y);
		draw3DFrame(show && focus, b.x, b.y,
			b.width - b.x, b.height - b.y);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}*/

	void recalibrateBounds(int width, int height, int depth) {
		boolean first = true;
		minCol = width / 2 - 1;
		maxCol = width / 2 + 1;
		minRow = height / 2 - 1;
		maxRow = height / 2 + 1;
		minStack = depth / 2 - 1;
		maxStack = depth / 2 + 1;

		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int minArea = width * height;
			int col = getColumn(position, width);
			int row = getRow(position, width, minArea);
			int stack = getStack(position, minArea);
			if (first) {
				minCol = maxCol = col;
				if (col > 0)
					minCol--;
				if (col < width - 1)
					maxCol++;
				minRow = maxRow = row;
				if (row > 0)
					minRow--;
				if (row < height - 1)
					minRow++;
				minStack = maxStack = stack;
				if (stack > 0)
					minStack--;
				if (stack < depth - 1)
					maxStack++;
				first = false;
			} else {
				if (col - 1 < minCol && col > 0)
					minCol = col - 1;
				if (col + 1 > maxCol && col < width - 1)
					maxCol = col + 1;
				if (row - 1 < minRow && row > 0)
					minRow = row - 1;
				if (row + 1 > maxRow && row < height - 1)
					maxRow = row + 1;
				if (stack - 1 < minStack && stack > 0)
					minStack = stack - 1;
				if (stack + 1 > maxStack && stack < depth - 1)
					maxStack = stack + 1;
			}
		}
	}

	boolean checkAliveBounds(int i, int j, int k) {
		boolean atEdge = false;

		if (minCol == 0 || maxCol == ncols - 1 ||
				i == 0 || i == ncols - 1) {
			minCol = 0;
			maxCol = ncols - 1;
			atEdge = true;
		} else if (i <= minCol) {
			minCol = i - 1;
		} else if (i >= maxCol) {
			maxCol = i + 1;
		}
		if (minRow == 0 || maxRow == nrows - 1 ||
				j == 0 || j == nrows - 1) {
			minRow = 0;
			maxRow = nrows - 1;
			atEdge = true;
		} else if (j <= minRow) {
			minRow = j - 1;
		} else if (j >= maxRow) {
			maxRow = j + 1;
		}
		if (minStack == 0 || maxStack == nstacks - 1 ||
				k == 0 || k == nstacks - 1) {
			minStack = 0;
			maxStack = nstacks - 1;
			atEdge = true;
		} else if (k <= minStack) {
			minStack = k - 1;
		} else if (k >= maxStack) {
			maxStack = k + 1;
		}
		return atEdge;
	}

	int scanField(String[] lineToken, int stringIndex,
			int offsetX, int offsetY, int offsetZ) {
		int polyhedron = getPolyhedron();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, k, n, p, startX, startY, startZ;
		String space = " ";
		int numNL = 0;

		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		startZ = offsetZ + nstacks / 2;
		if (!boundsCheck12(polyhedron, startX, startY, startZ)
				|| !boundsCheck14(polyhedron, startZ))
			startZ++;
		try {
			k = startZ;
			for (j = startY; ++m < lineToken.length; j++) {
				lineData = lineToken[m];
				if (lineData.length() == 0) {
					if (numNL > 0)
						break; // allows flexibility
					k++;
					j = startY - 1;
					numNL++;
					continue;
				}
				numNL = 0;
				n = 0;
				i = startX;
				while (n < lineData.length()) {
					byte oldCell, newCell;

					if (space.charAt(0) == lineData.charAt(n) ||
							i < 0 || i >= ncols ||
							j < 0 || j >= nrows ||
							k < 0 || k >= nstacks) {
						n++;
						continue;
					}
					p = getPosition(i, j, k);
					newCell = 0;
					if (cellCheck(polyhedron, i, j, k))
						newCell = (byte) ((Ant3DInterface.aliveChar(lineData.charAt(n))) ? 1 : 0);
					if (newCell != 0) {
						count++;
					}
					oldCell = field.getCell(p);
					if (oldCell != newCell) {
						field.setCell(p, newCell);
						CellInfo info = new CellInfo(
							p, newCell, (byte) 0);
						drawCell(info);
						if (newCell == 0) {
							nCells--;
						} else {
							nCells++;
							checkAliveBounds(i, j, k);
						}
						((Ant3DFrame) frame).updateCells(nCells);
					}
					i++;
					n++;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading (scanField)");
		}
		return count;
	}

	void drawField(boolean draw) {
		nCells = 0;
		if (draw) {
			showLiveCells();
			if ((Ant3DFrame) frame != null) {
				if (grid)
					drawGrid();
				((Ant3DFrame) frame).updateCells(nCells);
				capturer.show();
			}
		}
	}

	void redrawField() {
		eraseFrame();
		//drawFrame(true);
		drawField(true);
	}

	public Box getSelectedBox() {
		Point3D p = new Point3D();
		Dimension3D d = new Dimension3D();
		Rectangle rect = capturer.getRectangle();

		if (rect == null)
			return null;
		p.x = rect.x;
		p.y = rect.y;
		d.width = rect.width;
		d.height = rect.height;
		p.z = 0; // TODO: need a way to select in 3D
		d.depth = 1; // TODO: need a way to select in 3D
		return new Box(p, d);
	}

	void banner(String msg) {
		((Ant3DFrame) frame).displayMessage(msg);
	}

	Point3D inGrid(Point pt) {
		Graphics g = this.getGraphics();
		int polyhedron = getPolyhedron();
		int k = mouseDepth;

		if (g == null)
			return new Point3D(-1, -1, -1);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++)
		if (cellCheck(polyhedron, i, j, k) &&
			cellPl.inCell(g3d, pt, i, j, k)) {
			//System.out.println(i + " " + j + " " + k);
			return new Point3D(i, j, k);
		}
		return new Point3D(-1, -1, -1);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;
		int k = mouseDepth;
		int polyhedron = getPolyhedron();

		if (g == null)
			return;
		if (grid)
			color = Color.black;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++)
		if (cellCheck(polyhedron, i, j, k)) {
			cellPl.drawCell(g3d, wireframe, sphere,
				null, null, i, j, k);
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		//drawFrame(focus);
		drawField(true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	int getColumn(int position) {
		return position % this.ncols;
	}

	int getRow(int position) {
		return (position % this.area) / this.ncols;
	}

	int getStack(int position) {
		return position / this.area;
	}

	static int getColumn(int position, int ncols) {
		return position % ncols;
	}

	static int getRow(int position, int ncols, int area) {
		return (position % area) / ncols;
	}

	static int getStack(int position, int area) {
		return position / area;
	}

	int getPosition(int col, int row, int stack) {
		return getPosition(col, row, stack,
			this.ncols, this.area);
	}

	static int getPosition(int col, int row, int stack,
			int ncols, int area) {
		return stack * area + row * ncols + col;
	}

	int getColumn(CellInfo info) {
		return getColumn(info.getPosition());
	}

	int getRow(CellInfo info) {
		return getRow(info.getPosition());
	}

	int getStack(CellInfo info) {
		return getStack(info.getPosition());
	}

	void setPosition(CellInfo info, int col, int row, int stack) {
		info.setPosition(getPosition(col, row, stack));
	}

	Turmite3D heading(short direction, short gravity, short turn) {
		short oldDirection = direction;
		short newDirection = direction;
		short oldGravity = gravity;
		short newGravity = gravity;

		switch (neighbors) {
		case 12:
		    if (turn < 8) {
			if (turn == Ant3DRules.RTBS) {
			    if ((oldDirection & 1) == 1)
				newDirection = (short) (oldDirection - 1);
			    else
				newDirection = (short) (oldDirection + 1);
			} else if (turn < 6) {
			    newGravity = (short) Ant3DRules.newRGravityMap[turn][oldGravity][oldDirection % 4];
			    newDirection = (short) Ant3DRules.newRDirectionMap[turn][oldGravity][oldDirection % 4];
			}
		    }
		    break;
		case 8:
		    if (turn < 8) {
			if (turn == Ant3DRules.RTBS) {
			    if ((oldDirection & 1) == 1)
				newDirection = (short) (oldDirection - 1);
			    else
				newDirection = (short) (oldDirection + 1);
			} else if (turn < 6) {
			    newGravity = (short) Ant3DRules.newOGravityMap[turn][oldGravity][oldDirection];
			    newDirection = (short) Ant3DRules.newODirectionMap[turn][oldGravity][oldDirection];
			}
		    }
		    break;
		case 14:
		    if (turn < 6) {
			if (turn == Ant3DRules.TTBS) {
			    if ((oldDirection & 1) == 1)
				newDirection = (short) (oldDirection - 1);
			    else
				newDirection = (short) (oldDirection + 1);
			} else if (turn < 4) {
			    newGravity = (short) Ant3DRules.newTGravityMap[turn][oldGravity][oldDirection];
			    newDirection = (short) Ant3DRules.newTDirectionMap[turn][oldGravity][oldDirection];
			}
		    }
		    break;
		default:
		    if (turn < 6) {
			if (turn == Ant3DRules.TBS) {
			    if ((oldDirection & 1) == 1)
				newDirection = (short) (oldDirection - 1);
			    else
				newDirection = (short) (oldDirection + 1);
			} else if (turn < 4) {
			    newGravity = (short) Ant3DRules.newGravityMap[turn][oldGravity][oldDirection];
			    newDirection = (short) Ant3DRules.newDirectionMap[turn][oldGravity][oldDirection];
			}
		    }
		    break;
		}
		if (newGravity == Ant3DRules.NU)
			System.out.println("gravity corruption");
		if (newDirection == Ant3DRules.NU)
			System.out.println("direction corruption");
		if (debug) {
			System.out.println("oldDirection " +
				Ant3DRules.printDir(neighbors, oldDirection) +
				", oldGravity " +
				Ant3DRules.printDir((neighbors == 8) ? 8 : 6, oldGravity) +
				", turn " +
				Ant3DRules.printTurn(neighbors, turn) +
				", direction " +
				Ant3DRules.printDir(neighbors, newDirection) +
				", gravity " +
				Ant3DRules.printDir((neighbors == 8) ? 8 : 6, newGravity));
		}
		return new Turmite3D(0, 0, 0, newDirection, newGravity, (byte) 0);
	}

	int getNeighbors(int col, int row, int stack) {
		int n, p, count = 0;

		if (col == 0 || row == 0 || stack == 0 ||
				col == ncols - 1 || row == nrows - 1 ||
				stack == nstacks - 1) {
			edge = true;
			if (!wrap)
				return 0;
		}
		for (n = 0; n < Math.abs(neighbors); n++) {
			p = positionOfNeighbor(n, col, row, stack);
			try {
			if (field.getCell(p) != 0) {
				count++; // totalistic
			}
			} catch (ArrayIndexOutOfBoundsException e) {
				e.printStackTrace();
				System.out.println(n + " " + p + " "
						+ col + " " + row + " " + stack);
			}
		}
		return count;
	}

	int positionOfNeighbor(int cell, int i, int j, int k) {
		int col = i, row = j, stack = k, n = cell;

		// 6 faces
		// 8 corners
		// 12 edges, really rhombic dodecahedral
		// 14 truncated octahedronal
		// -14, alt 14 = faces + corners
		// 18 = edges + corners
		// 20 = faces + edges
		// 26 all
		/* faces */
		if (neighbors == 6 || neighbors == 14 || neighbors == -14 ||
				neighbors == 18 || neighbors == 26) {
			switch (n) {
			case 0:
				col = (col == 0) ? ncols - 1 : col - 1;
				break;
			case 1:
				col = (col + 1 == ncols) ? 0 : col + 1;
				break;
			case 2:
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 3:
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 4:
				if (neighbors == 14) {
					stack = (stack < 2) ? nstacks - 2 + stack : stack - 2;
				} else {
					stack = (stack == 0) ? nstacks - 1 : stack - 1;
				}
				break;
			case 5:
				if (neighbors == 14) {
					stack = (stack + 2 >= nstacks) ? stack + 2 - nstacks : stack + 2;
				} else {
					stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				}
				break;
			default:
				break;
			}
		} else {
			n += 6;
		}
		/* edges */
		if (neighbors == 12 || neighbors == 18 || neighbors == -18 ||
				neighbors == 20 || neighbors == 26) {
			switch (n) {
			case 6:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 7:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 8:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 9:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 10:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 11:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 12:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 13:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 14:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 15:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 16:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 17:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			default:
				break;
			}
		} else {
			n += 12;
		}
		/* corners */
		if (neighbors == 8 || neighbors == 14 || neighbors == -14 ||
				neighbors == 20 || neighbors == 26) {
			switch (n) {
			case 18:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				if (neighbors != 14 || (stack & 1) == 1)
				col = (col == 0) ? ncols - 1 : col - 1;
				if (neighbors != 14 || (stack & 1) == 1)
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 19:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				if (neighbors != 14 || (stack & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				if (neighbors != 14 || (stack & 1) == 1)
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 20:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				if (neighbors != 14 || (stack & 1) == 1)
				col = (col == 0) ? ncols - 1 : col - 1;
				if (neighbors != 14 || (stack & 1) == 0)
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 21:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				if (neighbors != 14 || (stack & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				if (neighbors != 14 || (stack & 1) == 0)
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 22:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				if (neighbors != 14 || (stack & 1) == 1)
				col = (col == 0) ? ncols - 1 : col - 1;
				if (neighbors != 14 || (stack & 1) == 1)
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 23:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				if (neighbors != 14 || (stack & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				if (neighbors != 14 || (stack & 1) == 1)
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 24:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				if (neighbors != 14 || (stack & 1) == 1)
				col = (col == 0) ? ncols - 1 : col - 1;
				if (neighbors != 14 || (stack & 1) == 0)
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 25:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				if (neighbors != 14 || (stack & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				if (neighbors != 14 || (stack & 1) == 0)
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			default:
				break;
			}
		} else {
			n += 8;
		}
		/* far corners */
		if (neighbors == -18) {
			switch (n) {
			case 26:
				col = (col < 2) ? ncols - 2 + col : col - 2;
				break;
			case 27:
				col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
				break;
			case 28:
				row = (row < 2) ? nrows - 2 + row : row - 2;
				break;
			case 29:
				row = (row + 2 >= nrows) ? row + 2 - nrows : row + 2;
				break;
			case 30:
				stack = (stack < 2) ? nstacks - 2 + stack : stack - 2;
				break;
			case 31:
				stack = (stack + 2 >= nstacks) ? stack + 2 - nstacks : stack + 2;
				break;
			default:
				break;
			}
		}
		return (getPosition(col, row, stack));
	}

	byte incCell(int col, int row, int stack) {
		int p = getPosition(col, row, stack);
		byte state = field.getCell(p);
		byte nextState = (byte) ((state < param.getNColors() - 1) ?
			state + 1 : 0);

		field.setCell(p, nextState);
		if (state < param.getNColors() - 1) {
			nCells++;
			checkAliveBounds(col, row, stack);
		} else {
			nCells--;
		}
		CellInfo info = new CellInfo(p, nextState, (byte) 0);
		drawCell(info);
		((Ant3DFrame) frame).updateCells(nCells);
		return nextState;
	}

	int cellCount() {
		return field.list.size();
	}

	boolean setCell(int col, int row, int stack,
			byte nextState) {
		int polyhedron = getPolyhedron();
		int p = getPosition(col, row, stack);
		byte state = field.getCell(p);

		if (state == 0 && nextState != 0) {
			field.setCell(p, nextState);
			if (!cellCheck(polyhedron, col, row, stack))
				throw new IllegalArgumentException("failed cell check on draw");
			nCells++;
			checkAliveBounds(col, row, stack);
		} else if (state == nextState) {
			return false;
		} else if (nextState == 0) { // state != 0
			field.setCell(p, (byte) 0);
			nCells--;
			return false;
		} else { // different sort of live cell
			field.setCell(p, nextState);
			cCells++;
		}
		/*if (draw) {
			CellInfo info = new CellInfo(p, nextState, false);

			drawCell(info);
			//showLiveCells();
			((Ant3DFrame) frame).updateCells(nCells);
		}*/
		return true;
	}

	/*void printState(int state) {
		CellList curr;
		int i = 0;

		curr = first[state]->next;
		System.out.println("state " + state);
		while (curr != list[state]) {
			System.out.println(i + ": position " +
				curr.info.getPosition() + ",	age " +
				curr.info.getAge() + ", state " +
				curr.info.getState() + ", toggle " +
				curr.info.getToggle());
			curr = curr.next;
			i++;
		}
	}*/

	static Color colorWheel(int state) {
		switch (state) {
		case 0: return Color.black;
		case 1: return Color.red;
		case 2: return Color.yellow;
		case 3: return Color.green;
		case 4: return Color.cyan;
		case 5: return Color.blue;
		case 6: return Color.magenta;
		case 7: return Color.pink; // orange and yellow too close
		default: return Color.white;
		}
	}

	void drawCell(CellInfo info) {
		Color color;
		int polyhedron = getPolyhedron();
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);
		byte state = info.getCell();

		if (!cellCheck(polyhedron, col, row, stack))
			return;
		if (state == 0) {
			color = background;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorWheel(state);
			}
		}
		if (g3d == null)
			return;
		if (g3d.getGraphics() == null) {
			/* can happen on static field */
			resizeField();
		}
		if (color != background) { // Changed for 3D
			cellPl.drawCell(g3d, wireframe, sphere,
				color, gridColor, col, row, stack);
		}
	}

	void drawCellMouth(CellInfo info, short direction) {
		Color color;
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);
		byte state = info.getCell();

		if (state == 0) {
			color = background;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorWheel(state);
			}
		}
		if (g3d == null)
			return;
		if (g3d.getGraphics() == null) {
			/* can happen on static field */
			resizeField();
		}
		if (!color.equals(background)) { // Changed for 3D
			cellPl.drawCellMouth(g3d, wireframe,
				color, gridColor, col, row, stack, direction);
		}
	}

	Ant3DMatrix getField() {
		return field;
	}

class Capturer {
	private Point3D anchor = new Point3D(0, 0, 0);
	private Point3D movable = new Point3D(0, 0, 0);
	private Point3D start = new Point3D(0, 0, 0);
	private Point3D stop = new Point3D(0, 0, 0);
	private int completed = -1;
	private final Color selectedColor = Color.magenta;

	Point3D calcPoint(int x, int y, int z) {
		Point3D t = new Point3D(x, y, z);

		if (t.x < 0)
			t.x = 0;
		if (t.y < 0)
			t.y = 0;
		if (t.z < 0)
			t.z = 0;
		if (t.x >= ncols)
			t.x = ncols - 1;
		if (t.y >= nrows)
			t.y = nrows - 1;
		if (t.z >= nstacks)
			t.z = nstacks - 1;
		return t;
	}

	void calcStartPoint() {
		if (anchor.x > movable.x)
			start.x = movable.x;
		else
			start.x = anchor.x;
		if (anchor.y > movable.y)
			start.y = movable.y;
		else
			start.y = anchor.y;
		if (anchor.z > movable.z)
			start.z = movable.z;
		else
			start.z = anchor.z;
	}

	void calcEndPoint() {
		if (anchor.x < movable.x)
			stop.x = movable.x;
		else
			stop.x = anchor.x;
		if (anchor.y < movable.y)
			stop.y = movable.y;
		else
			stop.y = anchor.y;
		if (anchor.z < movable.z)
			stop.z = movable.z;
		else
			stop.z = anchor.z;
	}

	void startCapture(int column, int row, int stack) {
		movable = anchor = calcPoint(column, row, stack);
		calcStartPoint();
		calcEndPoint();
		completed = 0;
		drawCaptureBox((grid) ? gridColor : background);
	}

	void continueCapture(int column, int row, int stack) {
		int captureColumns = getColumns();
		int captureRows = getRows();

		movable = calcPoint(column, row, stack);
		drawCaptureBox((grid) ? gridColor : background);
		if (column >= 0 && row >= 0 && stack >= 0 &&
				column < captureColumns && row < captureRows &&
				stack < stacks) {
			calcStartPoint();
			calcEndPoint();
			drawCaptureBox(selectedColor);
			return;
		}
		calcStartPoint();
		calcEndPoint();
	}

	void stopCapture() {
		if (completed == 0)
			completed = 1;
	}

	void clearCapture() {
		drawCaptureBox((grid) ? gridColor : background);
		anchor.setLocation(0, 0);
		movable.setLocation(0, 0);
		completed = -1;
	}

	private void drawCaptureBox(Color color) {
		int x = start.x;
		int y = start.y;
		int z = start.z;
		int width = stop.x - start.x + 1;
		int height = stop.y - start.y + 1;
		int depth = stop.z - start.z + 1;

		if (completed == -1)
			return;
		cellPl.drawCaptureBox(g3d, color,
			x, y, z, width, height, depth);
	}

	void show() {
		if (!anchor.equals(movable))
			drawCaptureBox(selectedColor);
	}

	Rectangle getRectangle() {
		if (completed == 1) {
			int x = start.x;
			int y = start.y;
			int width = stop.x - start.x + 1;
			int height = stop.y - start.y + 1;

			return new Rectangle(x, y, width, height);
		}
		return null;
	}

}
}
