package com.sillycycle.bagleyd.ant3d;

/*
 * @(#)Ant3DDialog.java
 *
 * Structure from Java Sourcebook by Ed Anuff, pp 287-289
 *
 * Copyright 2010 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.GridLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.Slider;

/**
 * The <code>Ant3DDialog</code> class holds the configuration
 * setup for the <code>Ant3DFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant3DDialog extends Dialog {
	private static final long serialVersionUID = 42L;
	static boolean active = false;
	static Ant3DFrame frame = null;

	static int columnsValue = Ant3DInterface.DEFAULT_COLUMNS;
	static int neighborIndexValue = Ant3DInterface.DEFAULT_NEIGHBOR_INDEX;
	static boolean randomOrientValue = false;
	static boolean lissajousValue = true;
	static boolean wrapValue = true;
	static boolean visibleValue = true;
	static boolean wireframeValue = false;
	static boolean useTableValue = false;
	static boolean mouthValue = true;

	Slider columnsChanger = new Slider("Field Size",
		columnsValue, Ant3DInterface.MIN_COLUMNS,
		Ant3DInterface.MAX_COLUMNS);
	Ant3DBoxes antBoxes = new Ant3DBoxes();
	JCheckBox randomOrientCheckBox = new JCheckBox("Random Orient");
	JCheckBox lissajousCheckBox = new JCheckBox("Lissajous");
	JCheckBox wrapCheckBox = new JCheckBox("Wrap");
	JCheckBox visibleCheckBox = new JCheckBox("Visible");
	JCheckBox wireframeCheckBox = new JCheckBox("Wireframe");
	JCheckBox useTableCheckBox = new JCheckBox("UseTable");
	JCheckBox mouthCheckBox = new JCheckBox("Mouth");

	public Ant3DDialog(Ant3DFrame frame, Frame parent,
			String titleText,
			String buttonText) {
		super(parent, titleText, true);

		JButton button = new JButton(buttonText);
		JPanel buttonPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel antBoxesPanel = new JPanel();
		JPanel togglePanel = new JPanel();

		columnsChanger.setWidth(150);
		buttonPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		sliderPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		sliderPanel.add(columnsChanger, BorderLayout.PAGE_START);
		buttonPanel.add(sliderPanel, BorderLayout.PAGE_START);
		antBoxesPanel.add(antBoxes, BorderLayout.PAGE_START);
		buttonPanel.add(antBoxesPanel, BorderLayout.PAGE_START);
		togglePanel.setLayout(new GridLayout(0, 2));
		randomOrientCheckBox.setSelected(randomOrientValue);
		togglePanel.add(randomOrientCheckBox);
		lissajousCheckBox.setSelected(lissajousValue);
		togglePanel.add(lissajousCheckBox);
		wrapCheckBox.setSelected(wrapValue);
		togglePanel.add(wrapCheckBox);
		visibleCheckBox.setSelected(visibleValue);
		togglePanel.add(visibleCheckBox);
		wireframeCheckBox.setSelected(wireframeValue);
		togglePanel.add(wireframeCheckBox);
		useTableCheckBox.setSelected(useTableValue);
		togglePanel.add(useTableCheckBox);
		mouthCheckBox.setSelected(mouthValue);
		togglePanel.add(mouthCheckBox);
		buttonPanel.add(togglePanel, BorderLayout.PAGE_START);

		buttonPanel.add(button, BorderLayout.PAGE_START);
		add(BorderLayout.PAGE_START, buttonPanel);
		pack();
		Ant3DDialog.active = true;
		Ant3DDialog.frame = frame;
		columnsChanger.addChangeListener(new ColumnsChangeListener());
		antBoxes.addActionListener(new AntActionListener());
		randomOrientCheckBox.addItemListener(new RandomOrientListener());
		lissajousCheckBox.addItemListener(new LissajousListener());
		wrapCheckBox.addItemListener(new WrapListener());
		visibleCheckBox.addItemListener(new VisibleListener());
		wireframeCheckBox.addItemListener(new WireframeListener());
		useTableCheckBox.addItemListener(new UseTableListener());
		mouthCheckBox.addItemListener(new MouthListener());

		button.addActionListener((ActionEvent event) -> {
			dispose();
			//else multiple Dialogs can be opened
			active = false;
		});

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});
	}

	class ColumnsChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = columnsValue;

			columnsValue = columnsChanger.getValue();
			if (oldValue != columnsValue) {
				frame.endGame();
				frame.setColumns(columnsValue);
				frame.startGame();
			}
		}
	}

	class AntActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = antBoxes.getNeighborIndex();
//System.out.println("actionPerformed: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class AntItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = antBoxes.getNeighborIndex();
//System.out.println("itemStateChanged: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class LissajousListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setLissajous(lissajousCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class RandomOrientListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setRandomOrient(randomOrientCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class WrapListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setWrap(wrapCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class VisibleListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setCellVisible(visibleCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class UseTableListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setUseTable(useTableCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class WireframeListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setWireframe(wireframeCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class MouthListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setMouth(mouthCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	public void setColumns(int columnsValue) {
		Ant3DDialog.columnsValue = columnsValue;
		columnsChanger.setValue(columnsValue);
	}

	public static int getColumns() {
		return columnsValue;
	}

	public void setNeighborIndex(int neighborIndexValue) {
		Ant3DDialog.neighborIndexValue = neighborIndexValue;
		antBoxes.setNeighborIndex(neighborIndexValue);
	}

	public static int getNeighborIndex() {
		return neighborIndexValue;
	}

	public void setRandomOrient(boolean flag) {
		Ant3DDialog.randomOrientValue = flag;
		randomOrientCheckBox.setSelected(flag);
	}

	public static boolean getRandomOrient() {
		return randomOrientValue;
	}

	public void setLissajous(boolean flag) {
		Ant3DDialog.lissajousValue = flag;
		lissajousCheckBox.setSelected(flag);
	}

	public static boolean getLissajous() {
		return lissajousValue;
	}

	public void setWrap(boolean flag) {
		Ant3DDialog.wrapValue = flag;
		wrapCheckBox.setSelected(flag);
	}

	public static boolean getWrap() {
		return wrapValue;
	}

	public void setCellVisible(boolean flag) {
		Ant3DDialog.visibleValue = flag;
		visibleCheckBox.setSelected(flag);
	}

	public static boolean getCellVisible() {
		return visibleValue;
	}

	public void setWireframe(boolean flag) {
		Ant3DDialog.wireframeValue = flag;
		wireframeCheckBox.setSelected(flag);
	}

	public static boolean getWireframe() {
		return wireframeValue;
	}

	public void setUseTable(boolean flag) {
		Ant3DDialog.useTableValue = flag;
		useTableCheckBox.setSelected(flag);
	}

	public static boolean getUseTable() {
		return useTableValue;
	}

	public void setMouth(boolean flag) {
		Ant3DDialog.mouthValue = flag;
		mouthCheckBox.setSelected(flag);
	}

	public static boolean getMouth() {
		return mouthValue;
	}

	public static boolean getActive() {
		return Ant3DDialog.active;
	}
}
