package com.sillycycle.bagleyd.ant3d;

/*
 * @(#)Ant3DRules.java
 *
 * Copyright 2010 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.ArrayList;
import java.util.HashMap;

import com.sillycycle.bagleyd.util.FiniteStateMachine;

/**
 * The <code>Ant3DRules</code> class handles the rules of life.
 * The way implemented should be good for 2 or 3 dimensional life.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant3DRules {
	static final int MAX_NEIGHBORS = 26; // 26 < 32 bits
	static final int BASE = 10;
	static final boolean DEBUG = false;
	static final int MIN_RULE = 18;
	static final int MAX_RULE = 65530;
	int ncolors = 8;
	int nstates = 1;

	private int neighborIndex;
	//private String ruleName;
	static int turingRule = MIN_RULE;

	public Ant3DRules(/*String ruleName*/) {
		//this.ruleName = ruleName;
	}

	/*public Ant3DRules(int neighbors, String ruleName) {
		this.neighborIndex = neighborToIndex(neighbors);
		//this.ruleName = ruleName;
		init();
	}*/

	public Ant3DRules(int neighbors) {
		this.neighborIndex = neighborToIndex(neighbors);
		init();
	}

	protected static String toString(int[] values) {
		StringBuilder string = new StringBuilder();

		for (int i = 0; i < values.length; i++) {
			if (i != 0)
				string.append(", ");
			string.append(i).append(": ").append(values[i]);
		}
		return string.toString();
	}

	public int getNeighbors() {
		return indexToNeighbor[neighborIndex];
	}

	public void setNeighborIndex(int neighborIndex) {
		this.neighborIndex = neighborIndex;
	}

	public void setNeighbors(int neighbors) {
		this.neighborIndex = neighborToIndex(neighbors);
	}

	public static String getRuleName(int neighborIndexValue) {
		return neighborIndexStrings[neighborIndexValue];
		//return Integer.toHexString(turingRule);
	}

	static final int[] neighborIndexToPolyhedron = {
		6, 6, 12, 14,
	};

	static final int MAX_NEIGHBOR_INDEXES = neighborIndexToPolyhedron.length;

	// Array of Generic Hash Tables
	ArrayList<HashMap<String, Integer>> arrayHashMap = new ArrayList<>();
//http://forums.sun.com/thread.jspa?threadID=530823
	final void init() {
		if (!arrayHashMap.isEmpty())
			arrayHashMap.clear();
		for (int i = 0; i < MAX_NEIGHBOR_INDEXES; i++) {
//System.out.println("Ant3DRules.init: " + " " + i);
			arrayHashMap.add(new HashMap<String, Integer>());
//System.out.println("Ant3DRules.init: " + " " + neighborIndex + " " + ant3dRules[neighborIndex].length);
		//for (int j = 0; j < ant3dRules[i].length; j++) {
//System.out.println("Ant3DRules.init: " + " " + indexToNeighbor[neighborIndex] + ant3dRules[neighborIndex][i].ruleName + " " + i);
		//	arrayHashMap.get(i).put(ant3dRules[i][j].ruleName, Integer.valueOf(j));
		//}
		}
	}

	static int neighborToPolyhedron(int neighbors) {
		switch (neighbors) {
		case 6: /* Cube */
		case 8:
		case -14: /* alt */
		case 18:
		case 20:
		case 26:
			return 6;
		case 12: /* Rhombic Dodecahedron, or Edges but every other cube */
		case -18:
			return 12;
		case 14: /* Truncated Octahedron, Bitruncated Cubic */
			return 14;
		default:
			System.out.println("no polyhedron known for " +
				neighbors + " neighbors");
			return 6;
		}
	}

	static final String[] neighborIndexStrings = {
		"6, Cube, von Neuman, Faces",
		"8, Cube, Corners",
		"12, Rhombic Dodecahedron, Kepler",
		"14, Truncated Octahedron, Kelvin (does not work, corkscrews)",
		/*"14, Cube, Faces and Corners",
		"18, Cube, Edges and Corners",
		"18, Rhombic Dodecahedron, Kepler",
		"20, Cube, Faces and Edges",
		"26, Cube, Moore"*/
	};

	static final int[] indexToNeighbor = {
		//6, 8, 12, 12, 14, -14, 18, -18, 20, 26
		6, 8, 12, 14,
	};

	static int neighborToIndex(int neighbors) {
		switch (neighbors) {
		case 6: /* Cube, von Neuman, Faces */
			return 0;
		case 8: /* Cube, Corners */
			return 1;
		case 12: /* Rhombic Dodecahedron, (every other cube, Edges) */
			return 2;
		case 14: /* Truncated Octahedron */
			return 3;
		//case -14: /* Cube, Faces and Corners */
		//	return 4;
		//case 18: /* Cube, Edges and Corners */
		//	return 5;
		//case -18: /* Rhombic Dodecahedron, (with corners) */
		//	return 6;
		//case 20: /* Cube, Faces and Edges */
		//	return 7;
		//case 26: /* Cube, Moore */
		//	return 8;
		default:
			System.out.println("no neighborhood like " +
				neighbors + " known");
			return Ant3DInterface.DEFAULT_NEIGHBOR_INDEX;
		}
	}

	static String parse(String rule) {
		return neighborIndexStrings[indexToNeighbor[parseNeighborFromRule(rule)]];
	}

	static int parseNeighborFromRule(String rule) {
		int neighbor = 0;
		int i = 0;
		boolean negative = false;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (letter >= '0' && letter <= '9')
				neighbor = neighbor * BASE + (letter - '0');
			else if (letter == '-')
				negative = !negative;
			else
				break;
		}
		if (negative)
			return neighborToIndex(-neighbor);
		return neighborToIndex(neighbor);
	}

	static final int STATES = 2;
	static final int MAXCOLORS = 8;

	/* Coordinate ant moves */
	public static final int NU = -1;
	public static final int XN = 0;
	public static final int XP = 1;
	public static final int YN = 2;
	public static final int YP = 3;
	public static final int ZN = 4;
	public static final int ZP = 5;
	public static final int ZNYNXN = 6; // Truncated Octahedron
	public static final int ZNYNXP = 7;
	public static final int ZNYPXN = 8;
	public static final int ZNYPXP = 9;
	public static final int ZPYNXN = 10;
	public static final int ZPYNXP = 11;
	public static final int ZPYPXN = 12;
	public static final int ZPYPXP = 13;
	public static final int XNYNZN = 0; // Corners
	public static final int XPYNZN = 1;
	public static final int XNYPZN = 2;
	public static final int XPYPZN = 3;
	public static final int XNYNZP = 4;
	public static final int XPYNZP = 5;
	public static final int XNYPZP = 6;
	public static final int XPYPZP = 7;
	public static final int XNYN = 0; // Rhombic Dodecahedron
	public static final int XPYN = 1;
	public static final int XNYP = 2;
	public static final int XPYP = 3;
	public static final int YNZN = 4;
	public static final int YPZN = 5;
	public static final int YNZP = 6;
	public static final int YPZP = 7;
	public static final int ZNXN = 8;
	public static final int ZPXN = 9;
	public static final int ZNXP = 10;
	public static final int ZPXP = 11;

	/* Relative ant moves */
	static final int TRS   = 0;  /* Turn right, then step */
	static final int TLS   = 1;  /* Turn left, then step */
	static final int TUS   = 2;  /* Turn up, then step */
	static final int TDS   = 3;  /* Turn down, then step */
	static final int FS    = 4;  /* Step */
	static final int TBS   = 5;  /* Turn back, then step */
	/* Uhhh for completeness I guess  there are some yaw moves and a 180 TU */
	static final int STR   = 6;  /* Step then turn right */
	static final int STL   = 7;  /* Step then turn left */
	static final int STU   = 8;  /* Step then turn up */
	static final int STD   = 9;  /* Step then turn down */
	static final int SF    = 10; /* Step */
	static final int STB   = 11; /* Step then turn back */
	/* Relative bee moves, leading R for Rhombic */
	static final int RTRS  = 0;  /* Turn right, then step */
	static final int RTLS  = 1;  /* Turn left, then step */
	/* can also include hard turns but did not yet */
	static final int RTURS = 2;  /* Turn up-right, then step */
	static final int RTDRS = 3;  /* Turn down-right, then step */
	static final int RTULS = 4;  /* Turn up-left, then step */
	static final int RTDLS = 5;  /* Turn down-left, then step */
	static final int RFS   = 6;  /* Step */
	static final int RTBS  = 7;  /* Turn back, then step */
	/* Uhhh for completeness I guess  there are some yaw moves and a 180 TU */
	static final int RSTR  = 8;  /* Step then turn right */
	static final int RSTL  = 9;  /* Step then turn left */
	static final int RSTUR = 10; /* Step then turn up-right */
	static final int RSTDR = 11; /* Step then turn down-right */
	static final int RSTUL = 12; /* Step then turn up-left */
	static final int RSTDL = 13; /* Step then turn down-left */
	static final int RSF   = 14; /* Step */
	/* Relative bubble moves, leading T for Truncated */
	static final int TTSURS  = 0;  /* Turn sharp up-right, then step */
	static final int TTSULS  = 1;  /* Turn sharp up-left, then step */
	static final int TTSDRS  = 2;  /* Turn sharp down-right, then step */
	static final int TTSDLS  = 3;  /* Turn sharp down-left, then step */
	static final int TTURS  = 4;  /* Turn up-right, then step */
	static final int TTULS  = 5;  /* Turn up-left, then step */
	static final int TTDRS  = 6;  /* Turn down-right, then step */
	static final int TTDLS  = 7;  /* Turn down-left, then step */
	static final int TFS   = 8;  /* Step */
	static final int TTBS  = 9;  /* Turn back, then step */

	// cube
	/*gravity XN, traveling along a direction YN, get TRS, now ZP */
	static int[][][] newDirectionMap = /*[4][6][6]*/ {
	{ /* TRS */
		{NU, NU, ZP, ZN, YN, YP}, /* oldGravity = XN */
		{NU, NU, ZN, ZP, YP, YN},
		{ZN, ZP, NU, NU, XP, XN}, /* oldGravity = YN */
		{ZP, ZN, NU, NU, XN, XP},
		{YP, YN, XN, XP, NU, NU}, /* oldGravity = ZN */
		{YN, YP, XP, XN, NU, NU}
	},
	{ /* TLS */
		{NU, NU, ZN, ZP, YP, YN}, /* oldGravity = XN */
		{NU, NU, ZP, ZN, YN, YP},
		{ZP, ZN, NU, NU, XN, XP}, /* oldGravity = YN */
		{ZN, ZP, NU, NU, XP, XN},
		{YN, YP, XP, XN, NU, NU}, /* oldGravity = ZN */
		{YP, YN, XN, XP, NU, NU}
	},
	{ /* TUS */
		{NU, NU, XP, XP, XP, XP}, /* oldGravity = XN */
		{NU, NU, XN, XN, XN, XN},
		{YP, YP, NU, NU, YP, YP}, /* oldGravity = YN */
		{YN, YN, NU, NU, YN, YN},
		{ZP, ZP, ZP, ZP, NU, NU}, /* oldGravity = ZN */
		{ZN, ZN, ZN, ZN, NU, NU}
	},
	{ /* TDS */
		{NU, NU, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU}
	}
	};
	/* turnDirection, oldGravity, oldDirection */
	static int[][][] newGravityMap = /*[4][6][6]*/ {
	{ /* TRS */
		{NU, NU, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU}
	},
	{ /* TLS */
		{NU, NU, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU}
	},
	{ /* TUS */
		{NU, NU, YN, YP, ZN, ZP}, /* oldGravity = XN */
		{NU, NU, YN, YP, ZN, ZP},
		{XN, XP, NU, NU, ZN, ZP}, /* oldGravity = YN */
		{XN, XP, NU, NU, ZN, ZP},
		{XN, XP, YN, YP, NU, NU}, /* oldGravity = ZN */
		{XN, XP, YN, YP, NU, NU}
	},
	{ /* TDS */
		{NU, NU, YP, YN, ZP, ZN}, /* oldGravity = XN */
		{NU, NU, YP, YN, ZP, ZN},
		{XP, XN, NU, NU, ZP, ZN}, /* oldGravity = YN */
		{XP, XN, NU, NU, ZP, ZN},
		{XP, XN, YP, YN, NU, NU}, /* oldGravity = ZN */
		{XP, XN, YP, YN, NU, NU}
	}
	};

	/*gravity XNYNZN, traveling along a direction XPYNZP, get TRS, now XPYPZP */
	static int[][][] newODirectionMap = /*[4][8][8]*/ {
	{ /* TSRS */
		{NU, XNYNZP, XPYNZN, XPYNZP, XNYPZN, XNYPZP, XPYPZN, NU}, /* oldGravity = XNYNZN */
		{XPYPZN, NU, XPYPZP, XPYNZP, XNYPZN, XNYNZN, NU, XNYNZP}, /* oldGravity = XPYNZN */
		{XNYPZP, XNYNZP, NU, XNYNZN, XPYPZP, NU, XPYPZN, XPYNZN}, /* oldGravity = XNYPZN */
		{XNYPZP, XNYPZN, XPYPZP, NU, NU, XNYNZN, XPYNZP, XPYNZN}, /* oldGravity = XPYPZN */
		{XPYNZP, XNYPZP, XPYPZP, NU, NU, XPYNZN, XNYNZN, XNYPZN}, /* oldGravity = XNYNZP */
		{XPYPZN, XPYPZP, NU, XNYPZP, XPYNZN, NU, XNYNZN, XNYNZP}, /* oldGravity = XPYNZP */
		{XPYNZP, NU, XNYNZP, XNYNZN, XPYPZP, XPYPZN, NU, XNYPZN}, /* oldGravity = XNYPZP */
		{NU, XNYPZN, XNYNZP, XNYPZP, XPYNZN, XPYPZN, XPYNZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TUS */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}, /* oldGravity = XNYNZN */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XPYNZN */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XNYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XPYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XNYNZP */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XPYNZP */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XNYPZP */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TSLS */
		{NU, XNYPZN, XNYNZP, XNYPZP, XPYNZN, XPYPZN, XPYNZP, NU}, /* oldGravity = XNYNZN */
		{XPYNZP, NU, XNYNZP, XNYNZN, XPYPZP, XPYPZN, NU, XNYPZN}, /* oldGravity = XPYNZN */
		{XPYPZN, XPYPZP, NU, XNYPZP, XPYNZN, NU, XNYNZN, XNYNZP}, /* oldGravity = XNYPZN */
		{XPYNZP, XNYPZP, XPYPZP, NU, NU, XPYNZN, XNYNZN, XNYPZN}, /* oldGravity = XPYPZN */
		{XNYPZP, XNYPZN, XPYPZP, NU, NU, XNYNZN, XPYNZP, XPYNZN}, /* oldGravity = XNYNZP */
		{XNYPZP, XNYNZP, NU, XNYNZN, XPYPZP, NU, XPYPZN, XPYNZN}, /* oldGravity = XPYNZP */
		{XPYPZN, NU, XPYPZP, XPYNZP, XNYPZN, XNYNZN, NU, XNYNZP}, /* oldGravity = XNYPZP */
		{NU, XNYNZP, XPYNZN, XPYNZP, XNYPZN, XNYPZP, XPYPZN, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TRS */
		{NU, XPYNZP, XPYPZN, XPYNZN, XNYPZP, XNYNZP, XNYPZN, NU}, /* oldGravity = XNYNZN */
		{XNYPZN, NU, XPYPZN, XPYPZP, XNYNZN, XNYNZP, NU, XPYNZP}, /* oldGravity = XPYNZN */
		{XNYNZP, XNYNZN, NU, XPYNZN, XNYPZP, NU, XPYPZP, XPYPZN}, /* oldGravity = XNYPZN */
		{XNYPZN, XNYNZN, XNYPZP, NU, NU, XPYNZN, XPYPZP, XPYNZP}, /* oldGravity = XPYPZN */
		{XPYNZN, XPYNZP, XNYNZN, NU, NU, XPYPZP, XNYPZN, XNYPZP}, /* oldGravity = XNYNZP */
		{XPYNZN, XPYPZN, NU, XPYPZP, XNYNZN, NU, XNYNZP, XNYPZP}, /* oldGravity = XPYNZP */
		{XNYNZP, NU, XNYNZN, XNYPZN, XPYNZP, XPYPZP, NU, XPYPZN}, /* oldGravity = XNYPZP */
		{NU, XPYPZN, XNYPZP, XNYPZN, XPYNZP, XPYNZN, XNYNZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TDS */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XPYNZN */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XNYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XPYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XNYNZP */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XPYNZP */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TLS */
		{NU, XPYPZN, XNYPZP, XNYPZN, XPYNZP, XPYNZN, XNYNZP, NU}, /* oldGravity = XNYNZN */
		{XNYNZP, NU, XNYNZN, XNYPZN, XPYNZP, XPYPZP, NU, XPYPZN}, /* oldGravity = XPYNZN */
		{XPYNZN, XPYPZN, NU, XPYPZP, XNYNZN, NU, XNYNZP, XNYPZP}, /* oldGravity = XNYPZN */
		{XPYNZN, XPYNZP, XNYNZN, NU, NU, XPYPZP, XNYPZN, XNYPZP}, /* oldGravity = XPYPZN */
		{XNYPZN, XNYNZN, XNYPZP, NU, NU, XPYNZN, XPYPZP, XPYNZP}, /* oldGravity = XNYNZP */
		{XNYNZP, XNYNZN, NU, XPYNZN, XNYPZP, NU, XPYPZP, XPYPZN}, /* oldGravity = XPYNZP */
		{XNYPZN, NU, XPYPZN, XPYPZP, XNYNZN, XNYNZP, NU, XPYNZP}, /* oldGravity = XNYPZP */
		{NU, XPYNZP, XPYPZN, XPYNZN, XNYPZP, XNYNZP, XNYPZN, NU}  /* oldGravity = XPYPZP */
	}
	};

	/* turnDirection, oldGravity, oldDirection */
	static int[][][] newOGravityMap = /*[4][8][8]*/ {
	{ /* TSRS */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XPYNZN */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XNYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XPYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XNYNZP */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XPYNZP */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TUS */
		{NU, XPYNZN, XNYPZN, XPYPZN, XNYNZP, XPYNZP, XNYPZP, NU}, /* oldGravity = XNYNZN */
		{XNYNZN, NU, XNYPZN, XPYPZN, XNYNZP, XPYNZP, NU, XPYPZP}, /* oldGravity = XPYNZN */
		{XNYNZN, XPYNZN, NU, XPYPZN, XNYNZP, NU, XNYPZP, XPYPZP}, /* oldGravity = XNYPZN */
		{XNYNZN, XPYNZN, XNYPZN, NU, NU, XPYNZP, XNYPZP, XPYPZP}, /* oldGravity = XPYPZN */
		{XNYNZN, XPYNZN, XNYPZN, NU, NU, XPYNZP, XNYPZP, XPYPZP}, /* oldGravity = XNYNZP */
		{XNYNZN, XPYNZN, NU, XPYPZN, XNYNZP, NU, XNYPZP, XPYPZP}, /* oldGravity = XPYNZP */
		{XNYNZN, NU, XNYPZN, XPYPZN, XNYNZP, XPYNZP, NU, XPYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYNZN, XNYPZN, XPYPZN, XNYNZP, XPYNZP, XNYPZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TSLS */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XPYNZN */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XNYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XPYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XNYNZP */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XPYNZP */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TRS */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XPYNZN */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XNYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XPYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XNYNZP */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XPYNZP */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TDS */
		{NU, XNYPZP, XPYNZP, XNYNZP, XPYPZN, XNYPZN, XPYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYPZP, NU, XPYNZP, XNYNZP, XPYPZN, XNYPZN, NU, XNYNZN}, /* oldGravity = XPYNZN */
		{XPYPZP, XNYPZP, NU, XNYNZP, XPYPZN, NU, XPYNZN, XNYNZN}, /* oldGravity = XNYPZN */
		{XPYPZP, XNYPZP, XPYNZP, NU, NU, XNYPZN, XPYNZN, XNYNZN}, /* oldGravity = XPYPZN */
		{XPYPZP, XNYPZP, XPYNZP, NU, NU, XNYPZN, XPYNZN, XNYNZN}, /* oldGravity = XNYNZP */
		{XPYPZP, XNYPZP, NU, XNYNZP, XPYPZN, NU, XPYNZN, XNYNZN}, /* oldGravity = XPYNZP */
		{XPYPZP, NU, XPYNZP, XNYNZP, XPYPZN, XNYPZN, NU, XNYNZN}, /* oldGravity = XNYPZP */
		{NU, XNYPZP, XPYNZP, XNYNZP, XPYPZN, XNYPZN, XPYNZN, NU}  /* oldGravity = XPYPZP */
	},
	{ /* TLS */
		{NU, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, XNYNZN, NU}, /* oldGravity = XNYNZN */
		{XPYNZN, NU, XPYNZN, XPYNZN, XPYNZN, XPYNZN, NU, XPYNZN}, /* oldGravity = XPYNZN */
		{XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN, NU, XNYPZN, XNYPZN}, /* oldGravity = XNYPZN */
		{XPYPZN, XPYPZN, XPYPZN, NU, NU, XPYPZN, XPYPZN, XPYPZN}, /* oldGravity = XPYPZN */
		{XNYNZP, XNYNZP, XNYNZP, NU, NU, XNYNZP, XNYNZP, XNYNZP}, /* oldGravity = XNYNZP */
		{XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP, NU, XPYNZP, XPYNZP}, /* oldGravity = XPYNZP */
		{XNYPZP, NU, XNYPZP, XNYPZP, XNYPZP, XNYPZP, NU, XNYPZP}, /* oldGravity = XNYPZP */
		{NU, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, XPYPZP, NU}  /* oldGravity = XPYPZP */
	}
	};

	// rhombic dodecahedron
	/* turnDirection, oldGravity, oldDirection */
	/* XNYN 0, XPYN 1, XNYP 2, XPYP 3
	   YNZN 4, YPZN 5, YNZP 6, YPZP 7
	   ZNXN 8, ZPXN 9, ZNXP 10,ZPXP 11 */
	/*gravity XN, traveling along a direction YNZN, get RTRS, now YNZP */
	static int[][][] newRDirectionMap = /*[6][6][4]*/ {
	{ /* RTRS:YNZN YPZN YNZP YPZP */
		{YPZN, YPZP, YNZN, YNZP}, /* YZ, oldGravity = XN */
		{YNZP, YNZN, YPZP, YPZN},
		{ZPXN, ZPXP, ZNXN, ZNXP}, /* ZX, oldGravity = YN */
		{ZNXP, ZNXN, ZPXP, ZPXN},
		{XPYN, XPYP, XNYN, XNYP}, /* XY, oldGravity = ZN */
		{XNYP, XNYN, XPYP, XPYN}
	},
	{ /* RTLS */
		{YNZP, YNZN, YPZP, YPZN}, /* oldGravity = XN */
		{YPZN, YPZP, YNZN, YNZP},
		{ZNXP, ZNXN, ZPXP, ZPXN}, /* oldGravity = YN */
		{ZPXN, ZPXP, ZNXN, ZNXP},
		{XNYP, XNYN, XPYP, XPYN}, /* oldGravity = ZN */
		{XPYN, XPYP, XNYN, XNYP}
	},
	{ /* RTURS */
		{ZNXN, XNYP, XNYN, ZPXN}, /* oldGravity = XN */
		{XPYN, ZNXP, ZPXP, XPYP},
		{XNYN, YNZP, YNZN, XPYN}, /* oldGravity = YN */
		{YPZN, XNYP, XPYP, YPZP},
		{YNZN, ZNXP, ZNXN, YPZN}, /* oldGravity = ZN */
		{ZPXN, YNZP, YPZP, ZPXP}
	},
	{ /* RTDRS */
		{ZNXP, XPYP, XPYN, ZPXP}, /* oldGravity = XN */
		{XNYN, ZNXN, ZPXN, XNYP},
		{XNYP, YPZP, YPZN, XPYP}, /* oldGravity = YN */
		{YNZN, XNYN, XPYN, YNZP},
		{YNZP, ZPXP, ZPXN, YPZP}, /* oldGravity = ZN */
		{ZNXN, YNZN, YPZN, ZNXP}
	},
	{ /* RTULS */
		{XNYN, ZNXN, ZPXN, XNYP}, /* oldGravity = XN */
		{ZNXP, XPYP, XPYN, ZPXP},
		{YNZN, XNYN, XPYN, YNZP}, /* oldGravity = YN */
		{XNYP, YPZP, YPZN, XPYP},
		{ZNXN, YNZN, YPZN, ZNXP}, /* oldGravity = ZN */
		{YNZP, ZPXP, ZPXN, YPZP}
	},
	{ /* RTDLS */
		{XPYN, ZNXP, ZPXP, XPYP}, /* oldGravity = XN */
		{ZNXN, XNYP, XNYN, ZPXN},
		{YPZN, XNYP, XPYP, YPZP}, /* oldGravity = YN */
		{XNYN, YNZP, YNZN, XPYN},
		{ZPXN, YNZP, YPZP, ZPXP}, /* oldGravity = ZN */
		{YNZN, ZNXP, ZNXN, YPZN}
	}
	};
	/* turnDirection, oldGravity, oldDirection */
	/* XNYN 0, XPYN 1, XNYP 2, XPYP 3
	   YNZN 4, YPZN 5, YNZP 6, YPZP 7
	   ZNXN 8, ZPXN 9, ZNXP 10,ZPXP 11 */
	static int[][][] newRGravityMap = /*[6][6][4]*/ {
	{ /* RTRS:YNZN YPZN YNZP YPZP */
		{XN, XN, XN, XN}, /* oldGravity = XN */
		{XP, XP, XP, XP},
		{YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP}
	},
	{ /* RTLS */
		{XN, XN, XN, XN}, /* oldGravity = XN */
		{XP, XP, XP, XP},
		{YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP}
	},
	{ /* RTURS */
		{YP, ZP, ZN, YN}, /* oldGravity = XN */
		{ZP, YN, YP, ZN},
		{ZP, XP, XN, ZN}, /* oldGravity = YN */
		{XP, ZN, ZP, XN},
		{XP, YP, YN, XN}, /* oldGravity = ZN */
		{YP, XN, XP, YN}
	},
	{ /* RTDRS */
		{YN, ZN, ZP, YP}, /* oldGravity = XN */
		{ZN, YP, YN, ZP},
		{ZN, XN, XP, ZP}, /* oldGravity = YN */
		{XN, ZP, ZN, XP},
		{XN, YN, YP, XP}, /* oldGravity = ZN */
		{YN, XP, XN, YP}
	},
	{ /* RTULS */
		{ZP, YN, YP, ZN}, /* oldGravity = XN */
		{YP, ZP, ZN, YN},
		{XP, ZN, ZP, XN}, /* oldGravity = YN */
		{ZP, XP, XN, ZN},
		{YP, XN, XP, YN}, /* oldGravity = ZN */
		{XP, YP, YN, XN}
	},
	{ /* RTDLS */
		{ZN, YP, YN, ZP}, /* oldGravity = XN */
		{YN, ZN, ZP, YP},
		{XN, ZP, ZN, XP}, /* oldGravity = YN */
		{ZN, XN, XP, ZP},
		{YN, XP, XN, YP}, /* oldGravity = ZN */
		{XN, YN, YP, XP}
	}
	};

	//truncated octahedron
	/* turnDirection, oldGravity, oldDirection */
	/* XN 0, XP 1, YN 2, YP 3 ZN 4, ZP 5,
	   ZNYNXN 6, ZNYNXP 7, ZNYPXN 8, ZNYPXP 9,
	   ZPYNXN 10, ZPYNXP 11, ZPYPXN 12, ZPYPXP 13 */
	/*gravity XN, traveling along a direction XNYNZN, get TTURS, now YN */
	static int[][][] newTDirectionMap = /*[14][6][8]*/ {
	{ /* TTSURS */
		{NU, NU, ZNYPXP, ZPYNXP, ZPYPXP, ZNYNXP, ZNYPXP, YP, ZPYPXP, ZP, ZNYNXP, ZN, ZPYNXP, YN}, /* oldGravity = XN */
		{NU, NU, ZPYPXN, ZNYNXN, ZPYNXN, ZNYPXN, ZP, ZPYNXN, YN, ZNYNXN, ZN, YP, ZPYPXN, ZNYPXN},
		{ZPYPXP, ZNYPXN, NU, NU, ZPYPXN, ZNYPXP, ZPYPXN, ZNYPXN, ZP, XN, ZPYPXP, ZNYPXP, XP, ZN}, /* oldGravity = YN */
		{ZNYNXP, ZPYNXN, NU, NU, ZPYNXP, ZNYNXN, XP, ZP, ZNYNXP, ZPYNXP, ZN, XN, ZNYNXN, ZPYNXN},
		{ZPYNXP, ZPYPXN, ZPYPXP, ZPYNXN, NU, NU, ZPYNXP, ZPYPXP, ZPYNXN, ZPYPXN, XP, YP, YN, XN}, /* oldGravity = ZN */
		{ZNYPXP, ZNYNXN, ZNYPXN, ZNYNXP, NU, NU, YP, XN, XP, YN, ZNYPXN, ZNYNXN, ZNYPXP, ZNYNXP}
	},
	{ /* TTSULS */
		{NU, NU, ZPYPXP, ZNYNXP, ZPYNXP, ZNYPXP, ZPYNXP, ZP, ZNYNXP, YN, ZPYPXP, YP, ZNYPXP, ZN}, /* oldGravity = XN */
		{NU, NU, ZNYPXN, ZPYNXN, ZPYPXN, ZNYNXN, YP, ZNYPXN, ZP, ZPYPXN, ZN, ZNYNXN, YN, ZPYNXN},
		{ZNYPXP, ZPYPXN, NU, NU, ZPYPXP, ZNYPXN, ZNYPXP, ZPYPXP, XP, ZP, ZNYPXN, ZPYPXN, ZN, XN}, /* oldGravity = YN */
		{ZPYNXP, ZNYNXN, NU, NU, ZPYNXN, ZNYNXP, ZP, XN, ZPYNXN, ZNYNXN, XP, ZN, ZPYNXP, ZNYNXP},
		{ZPYPXP, ZPYNXN, ZPYPXN, ZPYNXP, NU, NU, ZPYPXN, ZPYNXN, ZPYPXP, ZPYNXP, YP, XN, XP, YN}, /* oldGravity = ZN */
		{ZNYNXP, ZNYPXN, ZNYPXP, ZNYNXN, NU, NU, XP, YP, YN, XN, ZNYNXP, ZNYPXP, ZNYNXN, ZNYPXN}
	},
	{ /* TTSDRS */
		{NU, NU, ZNYPXN, ZPYNXN, ZPYPXN, ZNYNXN, YP, ZNYPXN, ZP, ZPYPXN, ZN, ZNYNXN, YN, ZPYNXN}, /* oldGravity = XN */
		{NU, NU, ZPYPXP, ZNYNXP, ZPYNXP, ZNYPXP, ZPYNXP, ZP, ZNYNXP, YN, ZPYPXP, YP, ZNYPXP, ZN},
		{ZPYNXP, ZNYNXN, NU, NU, ZPYNXN, ZNYNXP, ZP, XN, ZPYNXN, ZNYNXN, XP, ZN, ZPYNXP, ZNYNXP}, /* oldGravity = YN */
		{ZNYPXP, ZPYPXN, NU, NU, ZPYPXP, ZNYPXN, ZNYPXP, ZPYPXP, XP, ZP, ZNYPXN, ZPYPXN, ZN, XN},
		{ZNYNXP, ZNYPXN, ZNYPXP,ZNYNXN,  NU, NU, XP, YP, YN, ZN, ZNYNXP, ZNYPXP, ZNYNXN, ZNYPXN}, /* oldGravity = ZN */
		{ZPYPXP, ZPYNXN, ZPYPXN, ZPYNXP, NU, NU, ZPYPXN, ZPYNXN, ZPYPXP, ZPYNXP, YP, XN, XP, YN}
	},
	{ /* TTSDLS */
		{NU, NU, ZPYPXN, ZNYNXN, ZPYNXN, ZNYPXN, ZP, ZPYNXN, YN, ZNYNXN, YP, ZPYPXN, ZN, ZNYPXN}, /* oldGravity = XN */
		{NU, NU, ZNYPXP, ZPYNXP, ZPYPXP, ZNYNXP, ZNYPXP, YP, ZPYPXP, ZP, ZNYNXP, ZN, ZPYNXP, YN},
		{ZNYNXP, ZPYNXN, NU, NU, ZPYNXP, ZNYNXN, XP, ZP, ZNYNXP, ZPYNXP, ZN, XN, ZNYNXN, ZPYNXN}, /* oldGravity = YN */
		{ZPYPXP, ZNYPXN, NU, NU, ZPYPXN, ZNYPXP, ZPYPXN, ZNYPXN, ZP, XN, ZPYPXP, ZNYPXP, XP, ZN},
		{ZNYPXP, ZNYNXN, ZNYPXN, ZNYNXP, NU, NU, YP, YN, XP, YN, ZNYPXN, ZNYNXN, ZNYPXP, ZNYNXP}, /* oldGravity = ZN */
		{ZPYNXP, ZPYPXN, ZPYPXP, ZPYNXN, NU, NU, ZPYNXP, ZPYPXP, ZPYNXN, ZPYPXN, XP, YP, YN, XN}
	},
	{ /* TTURS */
		{NU, NU, ZPYNXP, ZNYPXP, ZNYNXP, ZPYPXP, YN, ZPYNXP, ZN, ZNYNXP, ZP, ZPYPXP, YP, ZNYPXP}, /* oldGravity = XN */
		{NU, NU, ZNYNXN, ZPYPXN, ZNYPXN, ZPYNXN, ZNYPXN, ZN, ZPYPXN, YP, ZNYNXN, YN, ZPYNXN, ZP},
		{ZNYPXN, ZPYPXP, NU, NU, ZNYPXP, ZPYPXN, ZN, XP, ZNYPXP, ZPYPXP, XN, ZP, ZNYPXN, ZPYPXN}, /* oldGravity = YN */
		{ZPYNXN, ZNYNXP, NU, NU, ZNYNXN, ZPYNXP, ZPYNXN, ZNYNXN, XN, ZN, ZPYNXP, ZNYNXP, ZP, XP},
		{ZPYPXN, ZPYNXP, ZPYNXN, ZPYPXP, NU, NU, XN, YN, YP, XP, ZPYPXN, ZPYNXN, ZPYPXP, ZPYNXP}, /* oldGravity = ZN */
		{ZNYNXN, ZNYPXP, ZNYNXP, ZNYPXN, NU, NU, ZNYNXP, ZNYPXP, ZNYNXN, ZNYPXN, YN, XP, XN, YP}
	},
	{ /* TTULS */
		{NU, NU, ZNYNXP, ZPYPXP, ZNYPXP, ZPYNXP, ZN, ZNYPXP, YP, ZPYPXP, YN, ZNYNXP, ZP, ZPYNXP}, /* oldGravity = XN */
		{NU, NU, ZPYNXN, ZNYPXN, ZNYNXN, ZPYPXN, ZPYNXN, YN, ZNYNXN, ZN, ZPYPXN, ZP, ZNYPXN, YP},
		{ZPYPXN, ZNYPXP, NU, NU, ZNYPXN, ZPYPXP, XN, ZN, ZPYPXN, ZNYPXN, ZP, XP, ZPYPXP, ZNYPXP}, /* oldGravity = YN */
		{ZNYNXN, ZPYNXP, NU, NU, ZNYNXP, ZPYNXN, ZNYNXP, ZPYNXP, ZN, XP, ZNYNXN, ZPYNXN, XN, ZP},
		{ZPYNXN, ZPYPXP, ZPYNXP, ZPYPXN, NU, NU, YN, XP, XN, YP, ZPYNXP, ZPYPXP, ZPYNXN, ZPYPXN}, /* oldGravity = ZN */
		{ZNYPXN, ZNYNXP, ZNYNXN, ZNYPXP, NU, NU, ZNYPXN, ZNYNXN, ZNYPXP, ZNYNXP, XN, YN, YP, XP}
	},
	{ /* TTDRS */
		{NU, NU, ZPYNXN, ZNYPXN, ZNYNXN, ZPYPXN, ZPYNXN, YN, ZNYNXN, ZN, ZPYPXN, ZP, ZNYPXN, YP}, /* oldGravity = XN */
		{NU, NU, ZNYNXP, ZPYPXP, ZNYPXP, ZPYNXP, ZN, ZNYPXP, YP, ZPYPXP, YN, ZNYNXP, ZP, ZPYNXP},
		{ZNYNXN, ZPYNXP, NU, NU, ZNYNXP, ZPYNXN, ZNYNXP, ZPYNXP, ZN, XP, ZNYNXN, ZPYNXN, XN, ZP}, /* oldGravity = YN */
		{ZPYPXN, ZNYPXP, NU, NU, ZNYPXN, ZPYPXP, XN, ZN, ZPYPXN, ZNYPXN, ZP, XP, ZPYPXP, ZNYPXP},
		{ZNYPXN, ZNYNXP, ZNYNXN, ZNYPXP, NU, NU, ZNYPXN, ZNYNXN, ZNYPXP, ZNYNXP, XN, YN, YP, XP}, /* oldGravity = ZN */
		{ZPYNXN, ZPYPXP, ZPYNXP, ZPYPXN, NU, NU, YN, XP, XN, YP, ZPYNXP, ZPYPXP, ZPYNXN, ZPYPXN}
	},
	{ /* TTDLS */
		{NU, NU, ZNYNXN, ZPYPXN, ZNYPXN, ZPYNXN, ZNYPXN, ZN, ZPYPXN, YP, ZNYNXN, YN, ZPYNXN, ZP}, /* oldGravity = XN */
		{NU, NU, ZPYNXP, ZNYPXP, ZNYNXP, ZPYPXP, YN, ZPYNXP, ZN, ZNYNXP, ZP, ZPYPXP, YP, ZNYPXP},
		{ZPYNXN, ZNYNXP, NU, NU, ZNYNXN, ZPYNXP, ZPYNXN, ZNYNXN, XN, ZN, ZPYNXP, ZNYNXP, ZP, XP}, /* oldGravity = YN */
		{ZNYPXN, ZPYPXP, NU, NU, ZNYPXP, ZPYPXN, ZN, XP, ZNYPXP, ZPYPXP, XN, ZP, ZNYPXN, ZPYPXN},
		{ZNYNXN, ZNYPXP, ZNYNXP, ZNYPXN, NU, NU, ZNYNXP, ZNYPXP, ZNYNXN, ZNYPXN, YN, XP, XN, YP}, /* oldGravity = ZN */
		{ZPYPXN, ZPYNXP, ZPYNXN, ZPYPXP, NU, NU, XN, YN, YP, XP, ZPYPXN, ZPYNXN, ZPYPXP, ZPYNXP}
	},
	};
	/* turnDirection, oldGravity, oldDirection */
	static int[][][] newTGravityMap = /*[14][6][8]*/ {
	{ /* TTSURS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTSULS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTSDRS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTSDLS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTURS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTULS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTDRS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	},
	{ /* TTDLS */
		{NU, NU, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN, XN}, /* oldGravity = XN */
		{NU, NU, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP, XP},
		{YN, YN, NU, NU, YN, YN, YN, YN, YN, YN, YN, YN, YN, YN}, /* oldGravity = YN */
		{YP, YP, NU, NU, YP, YP, YP, YP, YP, YP, YP, YP, YP, YP},
		{ZN, ZN, ZN, ZN, NU, NU, ZN, ZN, ZN, ZN, ZN, ZN, ZN, ZN}, /* oldGravity = ZN */
		{ZP, ZP, ZP, ZP, NU, NU, XP, XP, XP, XP, XP, XP, XP, XP}
	}
	};

	static final int[][] table = new int[][] {
	/* Generated */
	{	/* 0: RLU */
		3, 1,
		1, TRS, 0, 2, TLS, 0, 0, TUS, 0
	},
	{	/* 1: RUL */
		3, 1,
		1, TRS, 0, 2, TUS, 0, 0, TLS, 0
	},
	{	/* 2: RUD */
		3, 1,
		1, TRS, 0, 2, TUS, 0, 0, TDS, 0
	},
	{	/* 3: RLUU */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TUS, 0, 0, TUS, 0
	},
	{	/* 4: RRLU 32 */
		4, 1,
		1, TRS, 0, 2, TRS, 0, 3, TLS, 0, 0, TUS, 0
	},
	{	/* 5: RRUL ? */
		4, 1,
		1, TRS, 0, 2, TRS, 0, 3, TUS, 0, 0, TLS, 0
	},
	{	/* 6: RRUD 22 */
		4, 1,
		1, TRS, 0, 2, TRS, 0, 3, TUS, 0, 0, TDS, 0
	},
	{	/* 7: RLRU 188 */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TRS, 0, 0, TUS, 0
	},
	{	/* 8: RLLU ? */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TLS, 0, 0, TUS, 0
	},
	{	/* 9: RLUR 46 ! */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TUS, 0, 0, TRS, 0
	},
	{	/* 10: RLUL 32 */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TUS, 0, 0, TLS, 0
	},
	{	/* 11: RLUU 102 */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TUS, 0, 0, TUS, 0
	},
	{	/* 12: RLUD 28 */
		4, 1,
		1, TRS, 0, 2, TLS, 0, 3, TUS, 0, 0, TDS, 0
	},
	{	/* 13: RURL 30 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TRS, 0, 0, TLS, 0
	},
	{	/* 14: RURD ? */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TRS, 0, 0, TDS, 0
	},
	{	/* 15: RULL 22 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TLS, 0, 0, TLS, 0
	},
	{	/* 16: RULR 22 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TLS, 0, 0, TRS, 0
	},
	{	/* 17: RULU 22 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TLS, 0, 0, TUS, 0
	},
	{	/* 18: RUUL 26 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TUS, 0, 0, TLS, 0
	},
	{	/* 19: RULD 22 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TLS, 0, 0, TDS, 0
	},
	{	/* 20: RUUD 22 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TUS, 0, 0, TDS, 0
	},
	{	/* 21: RUDR 14 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TDS, 0, 0, TRS, 0
	},
	{	/* 22: RUDL ? */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TDS, 0, 0, TLS, 0
	},
	{	/* 23: RUDU 114 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TDS, 0, 0, TUS, 0
	},
	{	/* 24: RUDD 40 */
		4, 1,
		1, TRS, 0, 2, TUS, 0, 3, TDS, 0, 0, TDS, 0
	},
	{	/* 25: RRRULL ? */
		6, 1,
		1, TRS, 0, 2, TRS, 0, 3, TRS, 0, 4, TUS, 0,
		5, TLS, 0, 0, TLS, 0
	},
	{	/* 26: RLRUUUL 25436 */
		7, 1,
		1, TRS, 0, 2, TLS, 0, 3, TRS, 0, 4, TUS, 0,
		5, TUS, 0, 6, TUS, 0, 0, TLS, 0
	},
	{	/* 27: RRLDDDULRRLLLL * */
		14, 1,
		1, TRS, 0, 2, TRS, 0, 3, TLS, 0, 4, TDS, 0,
		5, TDS, 0, 6, TDS, 0, 7, TUS, 0, 8, TLS, 0,
		9, TRS, 0, 10, TRS, 0, 11, TLS, 0, 12, TLS, 0,
		13, TLS, 0, 0, TLS, 0
	}
	};

	static String printDir(int neighbors, int dir) {
		if (neighbors == 12) {
			switch (dir) {
			case XNYN: return "XNYN";
			case XPYN: return "XPYN";
			case XNYP: return "XNYP";
			case XPYP: return "XPYP";
			case YNZN: return "YNZN";
			case YPZN: return "YPZN";
			case YNZP: return "YNZP";
			case YPZP: return "YPZP";
			case ZNXN: return "ZNXN";
			case ZPXN: return "ZPXN";
			case ZNXP: return "ZNXP";
			case ZPXP: return "ZPXP";
			default: return "NUNU";
			}
		}
		if (neighbors == 8) {
			switch (dir) {
			case XNYNZN: return "XNYNZN";
			case XPYNZN: return "XPYNZN";
			case XNYPZN: return "XNYPZN";
			case XPYPZN: return "XPYPZN";
			case XNYNZP: return "XNYNZP";
			case XPYNZP: return "XPYNZP";
			case XNYPZP: return "XNYPZP";
			case XPYPZP: return "XPYPZP";
			default: return "NUNU";
			}
		}
		if (neighbors == 14) {
			switch (dir) {
			case ZNYNXN: return "ZNYNXN";
			case ZPYNXN: return "ZPYNXN";
			case ZNYPXN: return "ZNYPXN";
			case ZPYPXN: return "ZPYPXN";
			case ZNYNXP: return "ZNYNXP";
			case ZPYNXP: return "ZPYNXP";
			case ZNYPXP: return "ZNYPXP";
			case ZPYPXP: return "ZPYPXP";
			default: return "NUNU";
			}
		}
		switch (dir) {
		case XN: return "XN";
		case XP: return "XP";
		case YN: return "YN";
		case YP: return "YP";
		case ZN: return "ZN";
		case ZP: return "ZP";
		default: return "NU";
		}
	}

	static String printTurn(int neighbors, int dir) {
		if (neighbors == 12) {
			switch (dir) {
			case RTRS: return "RTRS";
			case RTLS: return "RTLS";
			case RTURS: return "RTURS";
			case RTDRS: return "RTDRS";
			case RTULS: return "RTULS";
			case RTDLS: return "RTDLS";
			default: return "RTXX";
			}
		}
		switch (dir) {
		case TRS: return "TRS";
		case TLS: return "TLS";
		case TUS: return "TUS";
		case TDS: return "TDS";
		default: return "TXX";
		}
	}

	FiniteStateMachine[] machine = null;

	void initMachine() {
		machine = new FiniteStateMachine[MAXCOLORS * STATES];
	}

	static boolean boring(int number) {
		int temp = number;
		boolean[] bucket = new boolean[4];

		bucket[0] = bucket[1] = bucket[2] = bucket[3] = false;

		while (temp >= 1) { // look for at least 3 buckets filled
			bucket[temp & 3] = true;
			temp >>= 2;
		}
		return ((!bucket[1] || !bucket[2] || !bucket[3]) &&
			(!bucket[0] || !bucket[2] || !bucket[3]) &&
			(!bucket[0] || !bucket[1] || !bucket[3]) &&
			(!bucket[0] || !bucket[1] || !bucket[2]));
	}

	private static int log4(int number) {
		int temp = number;
		int count = 0;

		while (temp >= 1) {
			temp >>= 2;
			count++;
		}
		return count;
	}

	void parseRule(int neighbors, int number, boolean useTable) {
		int newNumber = number;

		if (newNumber < MIN_RULE)
			newNumber = MIN_RULE;
		if (useTable) {
			newNumber -= MIN_RULE;
			newNumber = newNumber % table.length;
			getTable(/*neighbors,*/ newNumber);
			return;
		}
		/* if someone really insists let them have it */
		//while (boring(newNumber)) /* mod 4, need 3 out of 4 digits */
		//	newNumber--;*/
		getTurk(neighbors, log4(newNumber), newNumber);
	}

	void getTable(/*int neighbors,*/ int i) {
		ncolors = table[i][0];
		nstates = table[i][1];
		int total = ncolors * nstates;
		byte k = 2;

		for (int j = 0; j < total; j++) {
			byte color = (byte) table[i][k++];
			short turn = (short) table[i][k++];
			machine[j] = new FiniteStateMachine(color, turn, (byte) table[i][k++]);
		}
	}

	void getTurk(int neighbors, int logClass, int number) {
		int power4, j, total;
		StringBuilder rule = new StringBuilder();

		turingRule = number;
		power4 = 1 << ((logClass - 1) << 1);
		ncolors = logClass;
		nstates = 1;
		total = ncolors * nstates;
		for (j = 0; j < total; j++) {
			byte color = (byte) ((j + 1) % total);
			byte turn = (byte) (((power4 * 3) & number) / power4);

			rule.append(Integer.toString(turn));
			power4 >>= 2;
			machine[j] = new FiniteStateMachine(color, turn, (byte) 0);
			if (DEBUG)
				System.out.println("color " + color +
					", turn " + turn);
		}
		rule = new StringBuilder("0" + rule);
		if (DEBUG)
			System.out.println("neighbors " + neighbors +
				", Turk's number 0x" +
				Integer.toHexString(number) + " (" +
				rule.toString() + ")B4, colors " + ncolors);
	}

	// for debugging
	static String toString(int neighbors, /*Ant3DRules param,*/ boolean verbose) {
		StringBuilder sb = new StringBuilder("Rule");

		if (verbose)
			System.out.print("rule (" +
				neighbors + " neighborhood): ");
		//int neighborIndex = neighborToIndex(neighbors);
		if (verbose) {
			System.out.println(sb);
		}
		return sb.toString();
	}

	int getNColors() {
		return ncolors;
	}

	int getNStates() {
		return nstates;
	}
}
